use std::u16;

use crate::value::Value;

#[derive(Debug, PartialEq, Eq, strum_macros::Display)]
pub enum OptCode {
    OpReturn,
    OpConstant,
    OpNegate,
    OpAdd,
    OpSubstract,
    OpMultiply,
    OpDivide,
}
#[derive(Debug)]
pub struct Chunk {
    pub code: Vec<u8>,
    pub constants: Vec<Value>,
    pub lines: Vec<u16>,
}

impl From<OptCode> for u8 {
    fn from(code: OptCode) -> Self {
        match code {
            OptCode::OpReturn => 0,
            OptCode::OpConstant => 1,
            OptCode::OpNegate => 2,
            OptCode::OpAdd => 3,
            OptCode::OpSubstract => 4,
            OptCode::OpMultiply => 5,
            OptCode::OpDivide => 6,
        }
    }
}
#[derive(Debug)]
pub struct ConversionError {
    pub invalid_value: u8,
    pub message: String,
}

impl ConversionError {
    fn new(invalid_value: u8) -> Self {
        ConversionError {
            invalid_value,
            message: format!("Invalid opcode"),
        }
    }
}

impl TryFrom<u8> for OptCode {
    type Error = ConversionError;
    fn try_from(code: u8) -> Result<Self, Self::Error> {
        match code {
            0 => Ok(OptCode::OpReturn),
            1 => Ok(OptCode::OpConstant),
            2 => Ok(OptCode::OpNegate),
            3 => Ok(OptCode::OpAdd),
            4 => Ok(OptCode::OpSubstract),
            5 => Ok(OptCode::OpMultiply),
            6 => Ok(OptCode::OpDivide),
            _ => Err(ConversionError::new(code)),
        }
    }
}

impl Chunk {
    pub fn new() -> Self {
        Self {
            code: vec![],
            constants: vec![],
            lines: vec![],
        }
    }
    pub fn write(&mut self, byte: u8, line: u16) {
        self.code.push(byte);
        self.lines.push(line);
    }
    pub fn write_value(&mut self, add: Value) -> u8 {
        self.constants.push(add);
        return (self.constants.len() - 1).try_into().unwrap();
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn u8_to_optcode() {
        let pred: OptCode = 0.try_into().unwrap();
        assert_eq!(pred, OptCode::OpReturn);
        let pred: OptCode = 1.try_into().unwrap();
        assert_eq!(pred, OptCode::OpConstant);
    }
    #[test]
    fn basic_write() {
        let mut chunk = Chunk::new();
        let constant_idx: u8 = chunk.write_value(1.2);
        assert_eq!(constant_idx, 0);
        chunk.write(OptCode::OpConstant.into(), 123);
        chunk.write(constant_idx, 123);
        let pred: OptCode = chunk.code[0].try_into().unwrap();
        assert_eq!(pred, OptCode::OpConstant);
    }
    #[test]
    fn try_into_constant() {
        let mut chunk = Chunk::new();
        let constant_idx: u8 = chunk.write_value(1.2);
        chunk.write(OptCode::OpConstant.into(), 123);
        chunk.write(constant_idx, 123);
    }
}
