use peekmore::{PeekMore, PeekMoreIterator};
use std::fmt;
use std::str::Chars;

pub struct Scanner<'a> {
    source: &'a String,
    start: usize,
    chars: PeekMoreIterator<Chars<'a>>,
    current: usize,
    line: u16,
}

impl<'a> fmt::Display for Scanner<'a> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "{}:{}..{} '{}'",
            self.line,
            self.start,
            self.current,
            &self.source[self.start..self.current]
        )
    }
}

impl<'a> Scanner<'a> {
    pub fn new(source: &'a String) -> Self {
        Scanner {
            start: 0,
            current: 0,
            line: 1,
            chars: source.chars().peekmore(),
            source,
        }
    }
    pub fn compile(&mut self) {
        let mut line = 0;
        loop {
            let token = self.scan_token();

            if token.line != line {
                print!("{:4}", token.line);
                line = token.line;
            } else {
                print!("   |");
            }

            println!(
                ":{:<3} {:20} {:20}",
                token.start, token.token_type, token.lexeme
            );

            if let TokenType::TokenEof = token.token_type {
                break;
            }
        }
    }
    pub fn scan_token(&mut self) -> Token {
        self.skip_whitespace();
        self.start = self.current;
        match self.advance() {
            Some(c) => match c {
                c if c.is_digit(10) => return self.make_number(),
                c if c.is_ascii_alphabetic() || c == '_' => return self.make_identifier(),
                '(' => return self.make_token(TokenType::TokenLeftParen),
                ')' => return self.make_token(TokenType::TokenRightParen),
                '{' => return self.make_token(TokenType::TokenLeftBrace),
                '}' => return self.make_token(TokenType::TokenRightBrace),
                ';' => return self.make_token(TokenType::TokenSemicolon),
                ',' => return self.make_token(TokenType::TokenComma),
                '.' => return self.make_token(TokenType::TokenDot),
                '-' => return self.make_token(TokenType::TokenMinus),
                '+' => return self.make_token(TokenType::TokenPlus),
                '/' => return self.make_token(TokenType::TokenSlash),
                '*' => return self.make_token(TokenType::TokenStar),
                '!' => {
                    return self.make_token_if_matches(
                        '=',
                        TokenType::TokenBangEqual,
                        TokenType::TokenBang,
                    )
                }
                '=' => {
                    return self.make_token_if_matches(
                        '=',
                        TokenType::TokenEqualEqual,
                        TokenType::TokenEqual,
                    )
                }
                '<' => {
                    return self.make_token_if_matches(
                        '=',
                        TokenType::TokenLessEqual,
                        TokenType::TokenLess,
                    )
                }
                '>' => {
                    return self.make_token_if_matches(
                        '=',
                        TokenType::TokenGreaterEqual,
                        TokenType::TokenGreater,
                    )
                }
                '"' => return self.make_string(),
                _ => return self.make_error_token("Unexpected character."),
            },
            None => return self.make_eof(),
        };
    }
    fn make_identifier(&mut self) -> Token {
        while let Some(c) = self.chars.peek() {
            if c.is_alphabetic() || *c == '_' || c.is_digit(10) {
                self.advance();
            } else {
                break;
            }
        }
        let lexeme = &self.source[self.start..self.current];

        match lexeme {
            "and" => self.make_token(TokenType::TokenAnd),
            "class" => self.make_token(TokenType::TokenClass),
            "cap" => self.make_token(TokenType::TokenFalse),
            "else" => self.make_token(TokenType::TokenElse),
            "if" => self.make_token(TokenType::TokenIf),
            "nil" => self.make_token(TokenType::TokenNil),
            "no" => self.make_token(TokenType::TokenFalse),
            "or" => self.make_token(TokenType::TokenOr),
            "print" => self.make_token(TokenType::TokenPrint),
            "return" => self.make_token(TokenType::TokenReturn),
            "super" => self.make_token(TokenType::TokenSuper),
            "var" => self.make_token(TokenType::TokenVar),
            "while" => self.make_token(TokenType::TokenWhile),
            "false" => self.make_token(TokenType::TokenFalse),
            "for" => self.make_token(TokenType::TokenFor),
            "fun" => self.make_token(TokenType::TokenFun),
            "this" => self.make_token(TokenType::TokenThis),
            "true" => self.make_token(TokenType::TokenTrue),
            _ => return self.make_token(TokenType::TokenIdentifier),
        }
    }
    fn make_token_if_matches(
        &mut self,
        expected: char,
        on_match: TokenType,
        otherwise: TokenType,
    ) -> Token {
        if self.matches(expected) {
            self.make_token(on_match)
        } else {
            self.make_token(otherwise)
        }
    }
    fn matches(&mut self, expected: char) -> bool {
        match self.chars.peek() {
            Some(c) => {
                if c == &expected {
                    self.advance();
                    true
                } else {
                    false
                }
            }
            None => false,
        }
    }
    fn make_string(&mut self) -> Token {
        loop {
            match self.chars.peek() {
                Some(c) => match c {
                    '\n' => {
                        self.line += 1;
                        continue;
                    }
                    '"' => {
                        self.advance();
                        return self.make_token(TokenType::TokenString);
                    }
                    _ => {
                        self.advance();
                        continue;
                    }
                },
                None => return self.make_error_token("Unterminated string."),
            }
        }
    }
    fn make_number(&mut self) -> Token {
        while let Some(c) = self.peek() {
            if c.is_digit(10) {
                self.advance();
            } else {
                break;
            }
        }
        match (self.peek(), self.peek_next()) {
            (Some(c), Some(cc)) if c == '.' && cc.is_digit(10) => {
                self.advance();
                while let Some(c) = self.peek() {
                    if c.is_digit(10) {
                        self.advance();
                    } else {
                        break;
                    }
                }
                return self.make_token(TokenType::TokenNumber);
            }
            _ => {
                return self.make_token(TokenType::TokenNumber);
            }
        }
    }
    fn peek(&self) -> Option<char> {
        self.source.chars().nth(self.current)
    }
    fn peek_next(&self) -> Option<char> {
        if self.is_at_end() {
            return None;
        }
        return self.source.chars().nth(self.current + 1);
    }
    fn skip_whitespace(&mut self) {
        loop {
            match self.chars.peek() {
                None => return,
                Some(c) => match c {
                    ' ' | '\r' | '\t' => {
                        self.advance();
                        continue;
                    }
                    '\n' => {
                        self.line += 1;
                        self.advance();
                        continue;
                    }
                    '/' => match self.chars.peek_nth(1) {
                        Some(c) if *c == '/' => {
                            while let Some(c) = self.peek() {
                                if c == '\n' {
                                    self.line += 1;
                                    self.advance();
                                    return;
                                } else {
                                    self.advance();
                                }
                            }
                        }
                        _ => return,
                    },
                    _ => return,
                },
            }
        }
    }
    fn is_at_end(&self) -> bool {
        return self.current == self.source.len();
    }
    fn make_token(&self, token_type: TokenType) -> Token {
        Token {
            token_type,
            start: self.start,
            lexeme: self.source[self.start..self.current].to_string(),
            line: self.line,
        }
    }
    fn advance(&mut self) -> Option<char> {
        self.current += 1;
        self.chars.next()
    }
    fn make_error_token(&self, message: &str) -> Token {
        Token {
            token_type: TokenType::TokenError,
            start: self.start,
            lexeme: message.to_string(),
            line: self.line,
        }
    }
    fn make_eof(&self) -> Token {
        Token {
            token_type: TokenType::TokenEof,
            start: self.start,
            lexeme: "".to_string(),
            line: self.line,
        }
    }
}

pub struct Token {
    pub token_type: TokenType,
    pub start: usize,
    pub lexeme: String,
    pub line: u16,
}

#[derive(Clone, Copy, strum_macros::Display, Debug, PartialEq, Eq)]
pub enum TokenType {
    // Single-character tokens.
    TokenLeftParen,
    TokenRightParen,
    TokenLeftBrace,
    TokenRightBrace,
    TokenComma,
    TokenDot,
    TokenMinus,
    TokenPlus,
    TokenSemicolon,
    TokenSlash,
    TokenStar,
    // One or two character tokens.
    TokenBang,
    TokenBangEqual,
    TokenEqual,
    TokenEqualEqual,
    TokenGreater,
    TokenGreaterEqual,
    TokenLess,
    TokenLessEqual,
    // Literals.
    TokenIdentifier,
    TokenString,
    TokenNumber,
    // Keywords.
    TokenAnd,
    TokenClass,
    TokenElse,
    TokenFalse,
    TokenFor,
    TokenFun,
    TokenIf,
    TokenNil,
    TokenOr,
    TokenPrint,
    TokenReturn,
    TokenSuper,
    TokenThis,
    TokenTrue,
    TokenVar,
    TokenWhile,

    TokenError,
    TokenEof,
}

#[cfg(test)]
mod tests {
    use crate::scanner;
    use crate::scanner::TokenType;
    #[test]
    fn single_chars() {
        assert_token(String::from(""), TokenType::TokenEof);
        assert_token(String::from("("), TokenType::TokenLeftParen);
        assert_token(String::from("}"), TokenType::TokenRightBrace);
        assert_token(String::from("-"), TokenType::TokenMinus);
        assert_token(String::from("+"), TokenType::TokenPlus);
        assert_token(String::from("/"), TokenType::TokenSlash);
    }

    #[test]
    fn double_chars() {
        assert_token(String::from("=="), TokenType::TokenEqualEqual);
        assert_token(String::from("!="), TokenType::TokenBangEqual);
        assert_token(String::from(">"), TokenType::TokenGreater);
        assert_token(String::from(">="), TokenType::TokenGreaterEqual);
    }
    #[test]
    fn strings() {
        assert_token_lexeme(String::from("\"mox\""), TokenType::TokenString, "\"mox\"");
        assert_token_lexeme(String::from("\"\""), TokenType::TokenString, "\"\"");
    }

    #[test]
    fn numbers() {
        assert_token_lexeme(String::from("0"), TokenType::TokenNumber, "0");
        assert_token_lexeme(String::from("4"), TokenType::TokenNumber, "4");
        assert_token_lexeme(String::from("42"), TokenType::TokenNumber, "42");
        assert_token_lexeme(String::from("13.99"), TokenType::TokenNumber, "13.99");
    }
    #[test]
    fn newlines() {
        assert_tokens(
            String::from("+\n//comment\n-"),
            &vec![TokenType::TokenPlus, TokenType::TokenMinus],
        );
    }

    #[test]
    fn identifier() {
        assert_token(String::from("class"), TokenType::TokenClass);
        assert_token(String::from("if"), TokenType::TokenIf);
        assert_token(String::from("while"), TokenType::TokenWhile);
        assert_token(String::from("true"), TokenType::TokenTrue);
        assert_token(String::from("false"), TokenType::TokenFalse);

        assert_token(String::from("cap"), TokenType::TokenFalse);

        assert_token_lexeme(String::from("mox"), TokenType::TokenIdentifier, "mox");
    }
    fn assert_token(source: String, expected: scanner::TokenType) {
        let mut scanner = scanner::Scanner::new(&source);
        let token = scanner.scan_token();

        assert_eq!(token.token_type, expected);
    }

    fn assert_token_lexeme(
        source: String,
        expected_type: scanner::TokenType,
        expected_lexeme: &str,
    ) {
        let mut scanner = scanner::Scanner::new(&source);
        let token = scanner.scan_token();

        assert_eq!(token.token_type, expected_type);
        assert_eq!(token.lexeme, expected_lexeme);
    }
    fn assert_tokens(source: String, expected_tokens: &Vec<TokenType>) {
        let mut scanner = scanner::Scanner::new(&source);
        for expected in expected_tokens {
            let actual = scanner.scan_token();
            assert_eq!(actual.token_type, *expected);
        }

        assert_eq!(scanner.scan_token().token_type, TokenType::TokenEof);
    }
}
