use crate::chunk::{Chunk, ConversionError, OptCode};
use crate::compiler;
use crate::debug::{disassemble_instruction, print_value, trace_enabled};
use crate::value::Value;

pub struct VM<'a> {
    chunk: Option<&'a Chunk>,
    ip: usize,
    stack: Vec<Value>,
}

#[derive(Debug, PartialEq, Eq)]
pub enum InterpretResult {
    InterpretOk,
    InterpretCompileError,
    InterpretRuntimeError,
}

impl<'a> VM<'a> {
    pub fn new() -> Self {
        VM {
            chunk: None,
            ip: 0,
            stack: vec![],
        }
    }
    pub fn interpret_chunk(&mut self, chunk: &'a Chunk) -> InterpretResult {
        self.chunk = Some(chunk);
        self.ip = 0;
        self.run()
    }
    pub fn interpret(&mut self, source: &str, chunk: &'a mut Chunk) -> InterpretResult {
        match compiler::compile(source, &chunk) {
            false => {
                return InterpretResult::InterpretCompileError;
            }
            true => {
                self.chunk = Some(chunk);
                self.ip = 0;
                return self.run();
            }
        }
    }

    fn read_byte(&mut self) -> u8 {
        match self.chunk {
            Some(chunk) => {
                let result = chunk.code[self.ip];
                self.ip += 1;
                return result;
            }
            None => {
                panic!("could not read byte")
            }
        }
    }
    fn read_constant(&mut self) -> f64 {
        match self.chunk {
            Some(chunk) => chunk.constants[self.read_byte() as usize],
            None => {
                panic!("could not read constant")
            }
        }
    }
    fn run(&mut self) -> InterpretResult {
        while let Some(chunk) = self.chunk {
            if self.ip >= chunk.code.len() {
                break;
            }
            if trace_enabled() {
                print!("          ");
                println!("{:?}", self.stack);
                disassemble_instruction(&chunk, self.ip);
            }
            let byte: Result<OptCode, ConversionError> = self.read_byte().try_into();

            match byte {
                Ok(instruction) => match instruction {
                    OptCode::OpReturn => match self.stack.pop() {
                        Some(v) => {
                            print_value(v);
                            println!();
                            return InterpretResult::InterpretOk;
                        }
                        None => {
                            return InterpretResult::InterpretOk;
                        }
                    },
                    OptCode::OpConstant => {
                        let constant: Value = self.read_constant();
                        self.stack.push(constant);
                    }
                    OptCode::OpNegate => match self.stack.pop() {
                        Some(v) => {
                            self.stack.push(-v);
                        }
                        None => {
                            return InterpretResult::InterpretRuntimeError;
                        }
                    },
                    OptCode::OpAdd => match (self.stack.pop(), self.stack.pop()) {
                        (Some(b), Some(a)) => {
                            self.stack.push(a + b);
                        }
                        (_, _) => {
                            return InterpretResult::InterpretRuntimeError;
                        }
                    },
                    OptCode::OpSubstract => match (self.stack.pop(), self.stack.pop()) {
                        (Some(b), Some(a)) => {
                            self.stack.push(a - b);
                        }
                        (_, _) => {
                            return InterpretResult::InterpretRuntimeError;
                        }
                    },
                    OptCode::OpMultiply => match (self.stack.pop(), self.stack.pop()) {
                        (Some(b), Some(a)) => {
                            self.stack.push(a * b);
                        }
                        (_, _) => {
                            return InterpretResult::InterpretRuntimeError;
                        }
                    },
                    OptCode::OpDivide => match (self.stack.pop(), self.stack.pop()) {
                        (Some(b), Some(a)) => {
                            self.stack.push(a / b);
                        }
                        (_, _) => {
                            return InterpretResult::InterpretRuntimeError;
                        }
                    },
                },
                Err(_e) => {
                    return InterpretResult::InterpretRuntimeError;
                }
            }
        }
        return InterpretResult::InterpretRuntimeError;
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn unary_opts() {
        let mut chunk = Chunk::new();

        let constant_idx: u8 = chunk.write_value(1.2);
        chunk.write(OptCode::OpConstant.into(), 123);
        chunk.write(constant_idx, 123);
        chunk.write(OptCode::OpNegate.into(), 124);
        chunk.write(OptCode::OpReturn.into(), 125);

        let mut vm: VM = VM::new();
        let result: InterpretResult = vm.interpret_chunk(&chunk);
        assert_eq!(result, InterpretResult::InterpretOk);
    }
    #[test]
    fn binary_opts() {
        let mut chunk = Chunk::new();

        let constant_idx: u8 = chunk.write_value(1.2);
        chunk.write(OptCode::OpConstant.into(), 123);
        chunk.write(constant_idx, 123);

        let constant_idx: u8 = chunk.write_value(5.2);
        chunk.write(OptCode::OpConstant.into(), 125);
        chunk.write(constant_idx, 125);

        chunk.write(OptCode::OpAdd.into(), 126);

        let constant_idx: u8 = chunk.write_value(3.0);
        chunk.write(OptCode::OpConstant.into(), 127);
        chunk.write(constant_idx, 127);

        chunk.write(OptCode::OpDivide.into(), 128);

        chunk.write(OptCode::OpReturn.into(), 126);

        let mut vm: VM = VM::new();
        let result: InterpretResult = vm.interpret_chunk(&chunk);
        assert_eq!(result, InterpretResult::InterpretOk);
    }
}
